
#include "accessdb.h"
#include "util.h"
#include "torsf.h"
#include "compare.h"

enum compare_type{
	m_process = 1,
	m_inst,
	m_dep,
	m_invoke
};

char **commondira;
char **commondirb;
int commondirnum;

static int file_match, file_total;

static int getcommondirs(const char *va, const char *vb);
static int get_default_commondirs(const char *va, const char *vb);
static int read_commondirs(FILE *fp);
static void store_compare(const char *filenm, const char *va, const char *vb);
static void insert_match(enum compare_type, long oid_a, long oid_b, int degree);

static struct db_process * init_db_process(const char *);
static struct db_inst * init_db_inst(long);
static int db_process_compare(struct db_process *, struct db_process *);
static int db_inst_compare(struct db_inst *, struct db_inst *);
static void store_cmp_result(struct db_inst *);


static void init_compare_table()
{
	PGresult *res;
	char sql[256];

	sprintf(sql, "DROP TABLE BTV_COMPARE");
	res = PQexec(conn, sql);
	PQclear(res);
	
	sprintf(sql, "CREATE TABLE BTV_COMPARE \
			(CMP_TYPE INT, CMP_A INT, CMP_B INT, CMP_DEGREE INT)");
	res = PQexec(conn, sql);
	checkcreatestatus(res);
	PQclear(res);
}

void comparebtv()
{
	char *version_a = get_input_string
		("COMPARE: Build view A [e.g. pgsql-1.0]");
	char *version_b = get_input_string
		("COMPARE: Build view B [e.g. pgsql-1.0]");
	char *filename = get_input_string
		("COMPARE: Store the compare result to [e.g. pgsql.rsf]");

	struct db_process *ps_a, *ps_b, *cur;
	struct db_inst *inst_a, *inst_b;
	
	if (strcmp(version_a, version_b) == 0)
	{
		printf("Why you want to compare it with itself?\n");
		return;
	}

	printf("\nComparing BTV %s with BTV %s ... \n", version_a, version_b);

	init_compare_table();
	begintransaction();
	
	commondirnum = getcommondirs(version_a, version_b);
	if (commondirnum <= 0)
		printf("Can't compare the two build time view\n");
	else
	{
		int i;
		for (i = 0; i < commondirnum; i++)
			printf("%s\t%s\n", commondira[i], commondirb[i]);

		/* init match_stat */
		file_match = 0;
		file_total = 0;
	
		ps_a = init_db_process(version_a);
		ps_b = init_db_process(version_b);

		db_process_compare(ps_a, ps_b);

		for (cur = ps_a; cur; cur = cur->next)
		{
			if (cur->same_ps < 0) continue;
			inst_a = init_db_inst(cur->oid);
			inst_b = init_db_inst(cur->same_ps);
			file_match += db_inst_compare(inst_a, inst_b);
			store_cmp_result(inst_a);
			db_inst_free(inst_a);
			db_inst_free(inst_b);
		}

		/* print out statistics information */
		printf("Comparison statistics:\n");
		printf("Match %d out of %d targets in version. (%f)\n", 
				file_match * 2, file_total, 
				(double)file_match * 2 / file_total);
		/*printf("Match %d out of %d targets in version %s. (%f)\n", 
				file_match, graph_a->capacity, version_a, 
				(double)file_match/graph_a->capacity);
		printf("Match %d out of %d targets in version %s. (%f)\n", 
				file_match, graph_b->capacity, version_b, 
				(double)file_match/graph_b->capacity);*/
		store_compare(filename, version_a, version_b); 
		db_process_free(ps_a);
		db_process_free(ps_b);
	}

	endtransaction();
	
	printf("End of comparison\n");

	free (version_a);
	free (version_b);
}

static int getcommondirs(const char *va, const char *vb)
{
	FILE *fp;
	fp = fopen("commondir", "r");
	if (!fp)
		return get_default_commondirs(va, vb);
	else
		return read_commondirs(fp);
}

static int get_default_commondirs(const char *va, const char *vb)
{
	PGresult *res;
	char sql[256];
	char *cst_sql = "SELECT PS_DIR FROM BTV_MAKEPROC \
			 WHERE PS_VERSION ='%s' AND PS_MAKELEVEL=0";
       	
	/* get top directory of version A */
	sprintf(sql, cst_sql, va);
	res = PQexec(conn, sql);
	checkselectstatus(res);

	if (PQntuples(res) != 1)
	{
		printf("\nError: There is no build facts of version %s!\n", va);
		PQclear(res);
		return 0;
	}
	commondira = (char **)xmalloc(sizeof(char *));
	commondira[0]= (char *)xstrdup(PQgetvalue(res, 0, 0));
	PQclear(res);

	/* get top directory of version B */
	sprintf(sql, cst_sql, vb);
	res = PQexec(conn, sql);
	checkselectstatus(res);

	if (PQntuples(res) != 1)
	{
		printf("\nError: There is no build facts of version %s!\n", vb);
		PQclear(res);
		return 0;
	}
	commondirb = (char **)xmalloc(sizeof(char *));
	commondirb[0]= (char *)xstrdup(PQgetvalue(res, 0, 0));
	PQclear(res);

	return 1;
}

static int read_commondirs(FILE *fp)
{		
	int i, number;
	char dira[256], dirb[256];

	fscanf (fp, "%d", &number);
	commondira = (char **)xmalloc(sizeof(char *) * number);
	commondirb = (char **)xmalloc(sizeof(char *) * number);

	i = 0;
	while (i < number && fscanf(fp, "%s\t%s", dira, dirb) != EOF)
	{
		commondira[i] = (char *)xstrdup(dira);
		commondirb[i] = (char *)xstrdup(dirb);
		i ++;
	}

	while (i < number)
	{
		commondira[i] = 0;
		commondirb[i] = 0;
	}
	return number;
}

static int compare_dirs(const char *dir_a, const char *dir_b)
{
	int i;
	const char *ca, *cb;
	
	for (i = 0; i < commondirnum; i++)
	{
		if (strncmp(commondira[i], dir_a, strlen(commondira[i])) != 0)
			continue;
		if (strncmp(commondirb[i], dir_b, strlen(commondirb[i])) != 0)
			continue;

		ca = dir_a + strlen(commondira[i]);
		cb = dir_b + strlen(commondirb[i]);
		if (strcmp(ca, cb ) == 0) return 0;
	}
	return 1;
}

static void insert_match(enum compare_type t, long oid_a, long oid_b, int degree)
{
	PGresult *res;
	char sql[256];

	sprintf(sql, "INSERT INTO BTV_COMPARE VALUES (%d, %d, %d, %d)",
			t, oid_a, oid_b, degree);
	res = PQexec(conn, sql); 
	checkinsertstatus(res);
	PQclear(res);
}

static void torsf_compare(FILE *fp)
{
	char sql[512];
	PGresult *res;
	int ntup, degree;
	long aid, bid;

	sprintf(sql, "SELECT CMP_A, CMP_B, CMP_DEGREE FROM BTV_COMPARE");
	res = PQexec(conn, sql);
	checkselectstatus(res);

	for (ntup = 0; ntup < PQntuples(res); ntup++)
	{
		aid = atol(PQgetvalue(res, ntup, 0));
		bid = atol(PQgetvalue(res, ntup, 1));
		fprintf(fp, "SAME\t%d\t%d\n", aid, bid);
		
		degree = atoi(PQgetvalue(res, ntup, 2));
		if (degree < 0)
			fprintf(fp, "DIFF\t%d\t%d\n", aid, bid);
	}
}

void torsf_part_file(const char *bnm, FILE *fp)
{
        char sql[512];
        PGresult *res;
        int ntup;

	sprintf(sql, "SELECT BTV_FILEINST.OID, INST_NAME, INST_DIR,\
			     INST_ISGOAL, INST_DBLCLN\
			FROM BTV_FILEINST, BTV_MAKEPROC \
		       WHERE INST_PROC = BTV_MAKEPROC.OID \
			 AND PS_VERSION = '%s'", bnm);

	res = PQexec(conn, sql);
	checkselectstatus(res);
	
	for (ntup = 0; ntup < PQntuples(res); ntup++)
	{
		int fid = atol(PQgetvalue(res, ntup, 0));	
		fprintf(fp, "LAYER\t%d\t%s\n", fid, bnm);
		fprintf(fp, "FILE\t%d\t%s\n", fid, PQgetvalue(res, ntup, 1));
		fprintf(fp, "CONTAIN\t%s\t%d\n", PQgetvalue(res, ntup, 2), fid);
		
		if (atoi(PQgetvalue(res, ntup, 3)) != 0)
			fprintf(fp, "GOAL\t%d\t0\n", fid);

		if (atoi(PQgetvalue(res, ntup, 4)) != 0)
			fprintf(fp, "DBLCLN\t%d\t%d\n", fid,
					atoi(PQgetvalue(res, ntup, 4)));
	}
	
	PQclear(res);
}

static void store_compare(const char *filenm, const char *va, const char *vb)
{
        FILE *fp;

        fp = fopen(filenm, "w");
        if (!fp)
        {
                printf("Error: can not open %s !!!\n ", filenm);
                abort();
        }

	fprintf(fp, "LAYERNAME\t%s\t0\n", va);
	fprintf(fp, "LAYERNAME\t%s\t1\n", vb);

	torsf_dir(va, fp);
	torsf_part_file(va, fp);
	torsf_dependency(va, fp);
	torsf_call(va, fp);
	
	torsf_dir(vb, fp);
	torsf_part_file(vb, fp);
	torsf_dependency(vb, fp);
	torsf_call(vb, fp);
	
	torsf_compare(fp);

	fclose(fp);								
}

struct db_process * db_process_(long id, const char *d, struct db_process *h)
{
	struct db_process *res = (struct db_process *)xmalloc(sizeof(*res));
	res->oid = id;
	res->dir = xstrdup(d);
	res->next = h;
	res->same_ps = -1;
	res->goal = 0;
	return res;
}

void db_process_free(struct db_process *p)
{
	struct db_process *n;

	if (!p) return;
	n = p->next;
	free (p->dir);
	db_inst_free (p->goal);
	free (p);
	db_process_free (n);
}

struct db_inst * db_inst_(
		long id, const char *d, const char *n, 
		int dbl, struct db_inst *h)
{
	struct db_inst *res = (struct db_inst *)xmalloc(sizeof(*res));
	res->oid = id;
	res->dir = xstrdup (d);
	res->name = xstrdup (n);
	res->next = h;
	res->same_inst = -1;
	res->dblcln = dbl;
	return res;
}

void db_inst_free (struct db_inst *i)
{
	struct db_inst *n;
	if (!i) return;
	n = i->next;
	free (i->name);
	free (i->dir);
	free (i);
	db_inst_free (n);
}

struct db_process * init_db_process(const char *v)
{
	char *cst_sql = "SELECT BTV_MAKEPROC.OID, PS_DIR, \
				BTV_FILEINST.OID, INST_DIR, INST_NAME\
			   FROM BTV_MAKEPROC, BTV_FILEINST\
			  WHERE BTV_MAKEPROC.OID = INST_PROC\
			    AND INST_ISGOAL != 0 \
			    AND PS_VERSION = '%s'\
			  ORDER BY BTV_MAKEPROC.OID DESC";
	char sql[512];
	PGresult *res;
	int ntup;
	long old_oid, cur_oid;
	struct db_process * ret_ps = 0;
	
	sprintf(sql, cst_sql, v);
	res = PQexec(conn, sql);
	checkselectstatus(res);

	old_oid = -1;
	for (ntup = 0; ntup < PQntuples(res); ntup++)
	{
		cur_oid = atol(PQgetvalue(res, ntup, 0));
		if (cur_oid != old_oid)
		{
			ret_ps = db_process_(cur_oid, 
					PQgetvalue(res, ntup, 1), ret_ps);
			old_oid = cur_oid;
		}
		ret_ps->goal = db_inst_(atol(PQgetvalue(res, ntup, 2)),
				PQgetvalue(res, ntup, 3),
				PQgetvalue(res, ntup, 4),
				0,
				ret_ps->goal);
	}
	return ret_ps;
}

struct db_inst * init_db_inst(long proc_id)
{
	char *cst_sql= "SELECT OID, INST_DIR, INST_NAME, INST_DBLCLN\
			  FROM BTV_FILEINST \
			 WHERE INST_PROC = %d\
			 ORDER BY INST_NAME DESC";
	char sql [512];
	PGresult *res;
	struct db_inst *ret_inst = 0;
	int ntup;
	
	sprintf(sql, cst_sql, proc_id);
	res = PQexec(conn, sql);
	checkselectstatus(res);

	file_total += PQntuples(res);
	for (ntup = 0; ntup < PQntuples(res); ntup ++)
	{
		ret_inst = db_inst_(atol(PQgetvalue(res, ntup, 0)),
				PQgetvalue(res, ntup, 1),
				PQgetvalue(res, ntup, 2),
				atoi(PQgetvalue(res, ntup, 3)),
				ret_inst);
	}
	return ret_inst;
}

int db_process_compare(struct db_process *ps_a, struct db_process *ps_b)
{
	struct db_process *cur_a, *cur_b;
	int matched = 0, num_a = 0;
	
	for (cur_a = ps_a; cur_a; cur_a = cur_a->next)
	{
		num_a ++;
		for (cur_b = ps_b; cur_b; cur_b = cur_b->next)
		{
			if (cur_b->same_ps >= 0) continue;

			if (compare_dirs(cur_a->dir, cur_b->dir) != 0) continue;

			if (db_inst_compare(cur_a->goal, cur_b->goal) <= 0)
				continue;
			cur_a->same_ps = cur_b->oid;
			cur_b->same_ps = cur_a->oid;
			matched ++;
			break;
		}
	}
	printf("matched %d out of %d make process\n", matched, num_a);
	return matched;
}

int db_inst_compare(struct db_inst *inst_a, struct db_inst *inst_b)
{
	struct db_inst *cur_a, *cur_b;
	int matched = 0, cmp_res;
	
	for (cur_a = inst_a; cur_a; cur_a = cur_a->next)
	{
		for (cur_b = inst_b; cur_b; cur_b = cur_b->next)
		{
			if (cur_b->same_inst >= 0) continue;

			if (strcmp(cur_a->name, cur_b->name) != 0) continue;
			
			if (compare_dirs(cur_a->dir, cur_b->dir) != 0) continue;

			if (cur_a->dblcln != cur_b->dblcln) continue;
			
			cur_a->same_inst = cur_b->oid;
			cur_b->same_inst = cur_a->oid;
			matched ++;
			break;
		}
	}
	
	return matched;
}

void store_cmp_result(struct db_inst *inst)
{
	struct db_inst *cur;
	for(cur = inst; cur; cur=cur->next)
	{
		if (cur->same_inst >= 0)
		{
			insert_match(m_inst, cur->oid, cur->same_inst, 1);
		}
	}
}


