
#include "filedef.h"
#include "hash.h"
#include <string.h>
#include <assert.h>
#include "accessdb.h"

struct hash_table fileht;

static unsigned long file_hash_1 (const void *key)
{
	const struct file *f = (struct file *)key;
        return_STRING_HASH_1 (f->name);
}

static unsigned long file_hash_2 (const void *key)
{
	const struct file *f = (struct file *)key;
        return_STRING_HASH_2 (f->name);
}

static int file_hash_cmp (const void *x, const void *y)
{
	const struct file *fx = (struct file *)x;
	const struct file *fy = (struct file *)y;

	return !((strcmp(fx->name, fy->name)==0) && (strcmp(fx->dir, fy->dir)==0));
}

struct pidlist *_pidlist(struct pidlist *head, int id)
{
	struct pidlist *result = (struct pidlist *)xmalloc(sizeof(*result));
	result->pid = id;
	result->next = head;
	return result;
}

void pidlist_free(struct pidlist *ls)
{
	struct pidlist *head;

	while (ls)
	{
		head = ls;
		ls = ls->next;
		free (head);
	}
}

void pidlist_writeout(struct pidlist *ls)
{
	printf("cpids\n");
	while (ls)
	{
		printf("%d\n", ls->pid);
		ls = ls->next;
	}
}

struct file *_file(char *n, char *d)
{
	struct file *result =(struct file *)xmalloc(sizeof(*result));
	result->name = n;
	result->dir = d;
	result->isexported = 0;
	return result;
}

void file_free(struct file *f)
{
	free (f->name);
	free (f->dir);
	free (f);
}

void file_writeout(struct file *f)
{
	printf("file dir = %s, name = %s\n", f->dir, f->name);
}

struct filelist *_filelist(struct filelist *head, struct file *f)
{
	struct filelist *result = (struct filelist *)xmalloc(sizeof(*result));
	result->f = f;
	result->next = head;
	result->order = 0;
	return result;
}

void filelist_free(struct filelist *ls)
{
	/* not free file base */
	struct filelist *head;

	while (ls)
	{
		head = ls;
		ls = ls->next;
		free (head);
	}
}	

void filelist_writeout(struct filelist *ls)
{
	printf("goals/deps\n");
	while (ls)
	{
		file_writeout(ls->f);
		ls = ls->next;
	}
}

struct fileinstance *_fileinstance (struct file *f, struct file *g)
{
	struct fileinstance *result = (struct fileinstance *)xmalloc(sizeof(*result));
	result->base = f;
	result->finstid = 0;
	result->double_colon = 0;
	result->num_dblcln = 0;
	result->istarget = 0;
	result->isphony = 0;
	result->cmdstxt = 0;
	result->execmds = 0;
	result->cpids = 0;
	result->deps.ord = 0;
	result->goal = g;
	result->num_deps = 0;
	return result;
}

struct fileinstance *finst_add_dblcln(struct fileinstance *tmpl)
{
	struct fileinstance *result = _fileinstance(tmpl->base, tmpl->goal);
	tmpl->num_dblcln++;
	result->double_colon = tmpl->num_dblcln;
	tmpl->deps.dbl = _fileinstlist(tmpl->deps.dbl, result);
	tmpl->num_deps ++;
	tmpl->deps.dbl->order = tmpl->num_deps ++;
	return result;
}

void finst_free(struct fileinstance *fi)
{
	/* not free file base */
	free (fi->cmdstxt);
	free (fi->execmds);
	pidlist_free (fi->cpids);
	filelist_free (fi->deps.ord);
	free (fi);
}

void finst_writeout(struct fileinstance *fi)
{
	file_writeout(fi->base);
	if (fi->isphony) printf("\tphony");
	if (fi->istarget) printf("\ttarget");	
	if (fi->cmdstxt) printf("\ncmds:%s\n", fi->cmdstxt);
	if (fi->execmds) printf("execmds:\n%s\n", fi->execmds);
	pidlist_writeout(fi->cpids);
	/*
	filelist_writeout(fi->deps);
	*/
}

struct fileinstlist *_fileinstlist (struct fileinstlist *head, struct fileinstance *fi)
{
	struct fileinstlist *result = (struct fileinstlist *)xmalloc(sizeof(*result));
	result->finst = fi;
	result->next = head;
	result->order = 0;
	return result;
}

void flist_free(struct fileinstlist *ls)
{
	struct fileinstlist *head;

	while (ls)
	{
		head = ls;
		finst_free(ls->finst);
		ls = ls->next;
		free (head);
	}
}

void flist_writeout(struct fileinstlist *ls)
{
	while (ls)
	{
		finst_writeout(ls->finst);
		ls = ls->next;
	}
}


void finst_setcmdstxt(struct fileinstance *fi, char *content, int len)
{
	char *start = content;
	assert(fi);
	
	while (*start == '\t' || *start == ' ' || *start == '\n') start ++;

	if (len <= (start - content) ) return;
	
	if (!fi->cmdstxt)
		fi->cmdstxt = (char *)xstrncpy(start, len - (start - content));
	else {
		char *tmp1 = fi->cmdstxt;
		char *tmp2 = (char *)xstrncpy(start, len - (start - content));
		fi->cmdstxt = (char *)xstr3cat(tmp1, '\n', tmp2);
		free (tmp1);
		free (tmp2);
	}
}

void finst_setexecmds(struct fileinstance *fi, char *content, int len)
{
	char *start = content;
	assert(fi);
	
	while (*start == '\t' || *start == ' ' || *start == '\n') start ++;
	
	if (len <= (start - content) ) return;
	
	if (!fi->execmds)
		fi->execmds = (char *)xstrncpy(start, len - (start - content));
	else {
		char *tmp1 = fi->execmds;
		char *tmp2 = (char *)xstrncpy(start, len - (start - content));
		fi->execmds = (char *)xstr3cat(tmp1, '\n', tmp2);
		free (tmp1);
		free (tmp2);
	}
}

void finst_addcpid(struct fileinstance *fi, char *content, int len)
{
	char *start = content;
	assert(fi);
	while (*start == '\t' || *start == ' ' || *start == '\n') start ++;
	fi->cpids = _pidlist(fi->cpids, atoi(start));
}

void finst_adddep(struct fileinstance *fi, char *content, int len, const char *dir)
{
	char *start = content;
	char *name;
	struct file *f;
	
	assert(fi);
	while (*start == '\t' || *start == ' ' || *start == '\n') start ++;
	name = (char *)xstrncpy(start, len - (start - content));
	f = locatefile(name, dir);
	fi->deps.ord = _filelist(fi->deps.ord, f);
	fi->num_deps ++;
	fi->deps.ord->order = fi->num_deps;
	free (name);
}

struct file *locatefile(const char *n, const char *d)
{
	char *nm, *nmstart;
	struct file *f;
	struct fileinstance *fi;
	struct file **file_slot;
	
	/* Step 1: seperate directory information from file name */
	nmstart = strrchr(n, '/');
	/* n doesn't contain dir information */
	if (nmstart == 0) 
	{
		char *name = (char *)xstrdup(n);
		char *dir = (char *)xstrdup(d);
		f = _file(name, dir);
	}
	/* n contain directory starting from root directory */
	else if (*n == '/') 
	{
		char *name = (char *)xstrdup(nmstart + 1);
		char *dir = (char *)xstrncpy(n, nmstart - n);
		f = _file(name, dir);
	}
	/* otherwise */
	else
	{
		char *dir = (char *)xmalloc(strlen(n) + strlen(d) + 1);
		char *name;
		char *dir_cur; 
		const char *name_cur;
		
		strcpy(dir, d);
		dir_cur = dir + strlen(d);
		*dir_cur = '/';
		dir_cur ++;
		
		name_cur = n;
		while (name_cur < nmstart)
		{
			if (*name_cur != '.')
			{
				*dir_cur = *name_cur;
				dir_cur++;
				name_cur++;
			}
			
			else if (strncmp(name_cur, "../", 3) == 0)  
			{
				name_cur += 3;
				*(dir_cur - 1) = '\0';
				dir_cur = strrchr(dir, '/');
				assert(dir_cur);
				dir_cur++;
			}
			else if (strncmp(name_cur, "./", 2) == 0) 
				name_cur += 2;
			else {
				*dir_cur = *name_cur;
				dir_cur++;
				name_cur++;
			}
		}

		if (*(dir_cur - 1) == '/')
			*(dir_cur - 1) = '\0';
		else
			*dir_cur = '\0';

		if (nmstart < n + strlen(n) - 1)
			name = (char *)xstrdup(nmstart + 1);
		else
			name = (char *)xstrdup("_SPACE");
		f = _file(name, dir);
	}

	/* Step 2: Locate file use filename and its dir, 
	 * if it is not found, insert it */
	file_slot = (struct file **)hash_find_slot(&fileht, f);
	if (HASH_VACANT (*file_slot)) 
		hash_insert_at(&fileht, f, file_slot);
	else 
	{
		file_free(f);
		f = *file_slot;
	}
	return f;
}

void init_hashtable()
{
        hash_init (&fileht, 500, file_hash_1, file_hash_2, file_hash_cmp);
}

void free_hashtable()
{
	hash_map (&fileht, (hash_map_func_t)file_free);
	hash_free(&fileht, 0);
}

void export_file_table()
{
	hash_map (&fileht, (hash_map_func_t)export_file);
}
