
#include "intragraph.h"
#include "util.h"
#include "accessdb.h"

struct GraphNode * GraphNode_(long oid, char* d, char* n, int dbl, int isd, int ist, int isp, int isg)
{
	struct GraphNode* res = (struct GraphNode *)xmalloc(sizeof(*res));
	res->id = oid;
	res->dir = xstrdup(d);
	res->name = xstrdup(n);
	res->dblcln = dbl;
	res->isdblcln = (isd)?1:0;
	res->istarget = (ist)?1:0;
	res->isphony  = (isp)?1:0;
	res->isgoal   = (isg)?1:0;
	res->edges = 0;
	res->isvisited = 0;
	return res;
}

struct GraphEdge * GraphEdge_(
		struct GraphNode *t, struct GraphEdge *e)
{
	struct GraphEdge *res = (struct GraphEdge *)xmalloc(sizeof(*res));
	res->to = t;
	res->next = e;
	return res;
}

struct Graph * Graph_(int num)
{
	struct Graph *res = (struct Graph *)xmalloc(sizeof(*res));
	res->capacity = num;
	res->nodes = (struct GraphNode**)xmalloc(sizeof(struct GraphNode*)*num);
	res->startpoint = 0;
	return res;
}

void graphnode_free(struct GraphNode *n)
{
	if (!n) return;
	free(n->dir);
	free(n->name);
	graphedge_free(n->edges);
	free(n);
}

void graphedge_free(struct GraphEdge *e)
{
	struct GraphEdge *next;
	if (!e) return;
	next = e->next;
	free(e);
	graphedge_free(next);
}

void graph_free(struct Graph *g)
{
	int i;
	if (!g) return;
	for (i =0; i < g->capacity; i++)
	{
		graphnode_free(g->nodes[i]);
	}
	free(g);
}

static struct GraphNode * get_graphnode(struct Graph *g, long oid)
{
	int i, m, j;
	i = 0;
	j = g->capacity;
	while (i <= j) 
	{
		m = (i + j) / 2;
		if (g->nodes[m]->id < oid)
			i = m + 1;
		else if (g->nodes[m]->id > oid)
			j = m - 1;
		else
			return g->nodes[m];
	}
	return 0;
}

static void init_graphedge(struct Graph *g, const char *v)
{
	PGresult *res;
	int i;
	long cur_oid = 0, toid, soid;
	struct GraphNode * target, * source;
	char sql[512];
	char *cst_sql ="(SELECT DEP_TARGET, DEP_SOURCE\
			   FROM BTV_DEPENDENCY \
			  WHERE DEP_VERSION = '%s'\
			  ORDER BY DEP_TARGET )\
			 UNION \
			(SELECT CALL_CALLER, CALL_CALLEE\
			   FROM BTV_CALL \
			  WHERE CALL_VERSION = '%s'\
			  ORDER BY CALL_CALLER)";
	sprintf(sql, cst_sql, v, v);
	res = PQexec(conn, sql);
	checkselectstatus(res);

	for (i = 0; i < PQntuples(res); i++)
	{
		toid = atol(PQgetvalue(res, i, 0));
		soid = atol(PQgetvalue(res, i, 1));
		if (toid != cur_oid)
		{
			target = get_graphnode(g, toid);
			cur_oid = toid;
		}
		source = get_graphnode(g, soid);
		target->edges = GraphEdge_(source, target->edges);
	}
	PQclear(res);
}

static void init_startpoint(struct Graph *g, const char *v)
{
	PGresult *res;
	int i;
	char sql [512];
	char *cst_sql ="SELECT BTV_FILEINST.OID\
			  FROM BTV_MAKEPROC, BTV_FILEINST \
			 WHERE PS_VERSION = '%s' \
			   AND PS_MAKELEVEL = 0 \
			   AND BTV_MAKEPROC.OID = INST_PROC \
			   AND INST_ISGOAL != 0";
	struct GraphNode *goal;
	
	sprintf(sql, cst_sql, v);
	res = PQexec(conn, sql);
	checkselectstatus(res);

	for ( i = 0; i < PQntuples(res); i++)
	{
		goal = get_graphnode(g, atol(PQgetvalue(res, i, 0)));
		g->startpoint = GraphEdge_(goal, g->startpoint);
	}
	PQclear(res);
}

struct Graph* init_graph(const char * v)
{
	PGresult *res;
	int i;
	struct Graph * g;
	char sql[512];
        char *cst_sql = "SELECT OID, INST_DIR, INST_NAME, \
				INST_DBLCLN, INST_ISDBLCLN, \
				INST_ISTARGET, INST_ISPHONY, INST_ISGOAL\
			   FROM BTV_FILEINST \
			  WHERE INST_VERSION = '%s' \
			  ORDER BY OID";

	sprintf (sql, cst_sql, v);
	res = PQexec(conn, sql);
	checkselectstatus(res);

	g = Graph_(PQntuples(res));
	for ( i = 0; i < g->capacity; i++)
	{
		g->nodes[i] = GraphNode_(atol(PQgetvalue(res, i, 0)),
				PQgetvalue(res, i, 1),
				PQgetvalue(res, i, 2),
				atoi(PQgetvalue(res, i, 3)),
				atoi(PQgetvalue(res, i, 4)),
				atoi(PQgetvalue(res, i, 5)),
				atoi(PQgetvalue(res, i, 6)),
				atoi(PQgetvalue(res, i, 7)));
	}
	PQclear(res);
	init_startpoint(g, v);
	init_graphedge(g, v);
	return g;
}

