/* read.c
 * implement read.h */

#include <stdio.h>
#include "read.h"
#include "makeps.h"
#include "filedef.h"
#include "variable.h"
#include "util.h"

enum nametype{
	n_none,
	n_commandline,
	n_curdir,
	n_pid,
	n_ppid,
	n_variables,
	n_goal,
	n_file,
	n_doublecolon,
	n_target,
	n_phony,
	n_cmdstxt,
	n_execmds,
	n_cpids,
	n_dependencies
};

struct buffer
{
	char content[65536];
	int  len;
	char *cursor;
};

static struct buffer buf;

static FILE *logfile;
static long lineno;    /* for error report */

static struct makeprocess *curps = 0;
static struct fileinstance *curfi_tmpl = 0;
static struct fileinstance *curfi = 0;
static enum varsource curvs = o_invalid;
static struct file *curgoal = 0;

#define strneq(a, b, l) (strncmp ((a), (b), (l)) == 0)
#define streq(a, b)  (strcmp  ((a), (b)) == 0)

static enum nametype getname()
{
#define wordeq(s)  (strneq (s, buf.cursor, strlen(s)))
	if (*buf.cursor == '\t' || *buf.cursor == ' ') return n_none;
	
	if (wordeq("COMMANDLINE")) return n_commandline;
	if (wordeq("CURDIR")) return n_curdir;
	if (wordeq("PID")) return n_pid;
	if (wordeq("PPID")) return n_ppid;
	if (wordeq("VARIABLES")) return n_variables;
	if (wordeq("GOAL")) return n_goal;
	if (wordeq("FILE")) return n_file;
	if (wordeq("DOUBLECOLON")) return n_doublecolon;
	if (wordeq("TARGET")) return n_target;
	if (wordeq("PHONY")) return n_phony;
	if (wordeq("CMDSTXT")) return n_cmdstxt;
	if (wordeq("EXECMDS")) return n_execmds;
	if (wordeq("CPIDS")) return n_cpids;
	if (wordeq("DEPENDENCIES")) return n_dependencies;

	return -1;
#undef wordeq
}

#undef streq
#undef strneq

static void initbuffer()
{
	buf.len = 0;
	buf.cursor = buf.content;
}

static void die (int errorcode, const char *msg) 
{ 
	switch (errorcode) 
	{ 
		case 0: printf("ERROR F000: meet error at line %d, %s\n", lineno, msg);
			break;
		case 1: printf("ERROR F001: Can't open file %s\n", msg);
			break;
		case 2: printf("ERROR F002: line %d is too long\n", lineno);
			break;
		case 3: printf("ERROR F003: Log file is incomplete at line %d\n", lineno);
			break;
		default: break;
	}
}

/* return the length of the next line */
int getnextline(int oldlen)
{
	int i;
	int linelen = 0;
	char *cur;

	buf.cursor += oldlen;
	/* skip empty line */
	while (*buf.cursor == '\n' && buf.cursor < buf.content + buf.len)
	{
		buf.cursor ++;
		lineno++;
	}
	
	cur = buf.cursor;
	while (cur < buf.content + buf.len)
	{
		if (*cur == '\n')
		{
			lineno++;
			return linelen;
		}

		cur ++;
		linelen ++;
	}
	
	if ((buf.len > 0) && (buf.cursor == buf.content)) 
	{
		die(2, "");
		return -1;
	}
	
	/* there is nothing more in the logfile */
	if ((buf.len > 0 ) && (buf.len < sizeof(buf.content)))
	{
		if (linelen == 0) return -1;	/* EOF */
		die(3, "");
		return -1;
	}

	for (i = 0; i < (buf.content + buf.len - buf.cursor); i++)
	{
		buf.content[i] = buf.cursor[i];
	}
	
	buf.len = i;
	buf.len += fread((char *)(buf.content + buf.len), 1, 
			sizeof(buf.content) - buf.len, logfile);
	buf.cursor = buf.content;

	return getnextline(0);
}

/* load log from FILENM with build name BNM
 * return 0 if load successfully
 * otherwise return error code */
int load(const char *filenm, const char *bnm)
{
	int linelen;
	enum nametype lastname, curname;
	char *value;

	lineno = 0;

	logfile = fopen(filenm, "r");
	if (!logfile) {
		die (1, filenm);
		return -1;
	}

	initbuffer();
	
	lastname = n_none;
	
	linelen = getnextline(0);
	while (linelen >= 0)
	{
		curname = getname();
		switch (curname)
		{
		case n_commandline:
			if (curps) {
				export2db(curps);
				ps_free(curps);
				curps = 0;
			}
			/*new make process*/;
			curps = _makeprocess(bnm);
		case n_curdir:
			lastname = curname;
			break;
		case n_target:
			curfi->istarget = 1;
			break;
		case n_phony:
			curfi->isphony = 1;
			break;
		case n_doublecolon:
			curfi = finst_add_dblcln(curfi_tmpl);
			curps->files = _fileinstlist(curps->files, curfi);
			break;
		case n_pid:
		case n_ppid:
		case n_variables:
		case n_file:
		case n_cmdstxt:
		case n_execmds:
		case n_goal:
		case n_cpids:
		case n_dependencies:
			lastname = curname;
			break;
		case n_none:
			switch (lastname)
			{
			case n_commandline:
				ps_setcommandline(curps, buf.cursor, linelen);
			case n_curdir: 
				ps_setcurdir(curps, buf.cursor, linelen);
				break;
			case n_pid:
				ps_setpid(curps, buf.cursor, linelen);
				break;
			case n_ppid:
				ps_setppid(curps, buf.cursor, linelen);
				break;
			case n_variables:
				curvs = ps_addvariable(curps, buf.cursor, linelen, curvs);
				break;
			case n_goal:
				curgoal = ps_addgoal(curps,buf.cursor,linelen);
				break;
			case n_file:
				curfi = ps_addfileinst(curps, curgoal, buf.cursor, linelen);
				curfi_tmpl = curfi;
				break;
			case n_target:
			case n_phony:
				break;
			case n_cmdstxt:
				finst_setcmdstxt(curfi, buf.cursor, linelen);
				break;
			case n_execmds:
				finst_setexecmds(curfi, buf.cursor, linelen);
				break;
			case n_cpids:
				finst_addcpid(curfi, buf.cursor, linelen);
				break;
			case n_dependencies:
				finst_adddep(curfi, buf.cursor, linelen, curps->curdir);
				break;
			default: die(0, "unexpected name type");
				 return -1;
			}
			break;
		default: die(0, "unexpected name type");
			 return -1;
		}
		linelen = getnextline(linelen + 1);
	}
	
	if (curps) {
		export2db(curps);
		ps_free(curps);
		curps = 0;
	}

	fclose (logfile);
}

int loadlog()
{
	char *filenm = get_input_string
		("LOAD: Please input your log filename [e.g. makeoutput]");
	char *bnm = get_input_string
		("LOAD: Please input system building name [e.g. pgsql]");

	if (isloaded(bnm))
	{
		printf("Do nothing. \n");
		return;
	}
	
	printf("\nStart to loading %s %s ... \n\n", filenm, bnm);
	load(filenm, bnm);
	printf("\nEnd of load %s %s ... \n\n", filenm, bnm);

	free (filenm);
	free (bnm);
}

void remove_log()
{
	char *bnm = get_input_string
		("REMOVE: Please input system building name [e.g. pgsql-1.0]");
	removelog(bnm);
}
