#include <stdio.h>
#include <assert.h>

#include "include/rsf2dot.h"

struct hash_table graphNodes;
int start_printing = 0;
GraphEdge_ *edge_head = 0;
char *root_directory = 0;
char *layer_name[2] = {0, 0};

static unsigned long graph_hash_1 (const void *key)
{
	return_STRING_HASH_1 (((GraphNode_ *) key)->name);
}

static unsigned long graph_hash_2 (const void *key)
{
	return_STRING_HASH_2 (((GraphNode_ *) key)->name);
}

static int graph_hash_cmp (const void *x, const void *y)
{
	const GraphNode_ *gx = (GraphNode_ *)x;
	const GraphNode_ *gy = (GraphNode_ *)y;
	
	return_STRING_COMPARE(((GraphNode_ *)x)->name, ((GraphNode_ *)y)->name);
}

static int free_graphNode(GraphNode_ *n)
{
	free (n->name);
	if (n->layer) free (n->layer);

	if ((n->type == NT_FILE) && (n->filename)) free (n->filename);
	free (n);
}

static void free_hashtable()
{
	hash_map(&graphNodes, (hash_map_func_t)free_graphNode);
	hash_free(&graphNodes, 0);
}

void *xmalloc(unsigned int len)
{
	void *p = (void *)malloc(len);
	assert(p);
	return p;
}

static char *xstrdup(const char * src)
{
	int len;
	char *des;

	len = strlen(src);
	des = (char *)xmalloc(sizeof(char) * (len + 1));
	strcpy(des, src);
	des[len] = 0;
	return des;
}

static void xstrncpy(char *des, const char *src, int size)
{
	strncpy(des, src, size);
	des[size] = 0;
}

static void print_usage()
{
	printf("Usage: rsf2dot rsffilename [output dot filename]\n");
}

GraphNode_ * GraphNode(NodeType_ nt, const char * nm)
{
	GraphNode_ * node = (GraphNode_*) xmalloc(sizeof(*node));
	node->type = nt;
	node->name = xstrdup(nm);

	node->filename = (char *)strrchr(node->name, '/');
	if (node->filename) 
	{
		if (strlen(node->filename) > 1)	node->filename ++;
	}	
	
	node->isphony = 0;
	node->istarget = 0;
	node->isgoal = 0;
	node->isexecuted = 0;
	node->dblcln = -1;
	node->layer = 0;
	node->next = 0;
	node->children = 0;
	node->same = 0;
	node->isdup= 0;
	node->isdiff = 0;
	return node;
}

GraphEdge_ * GraphEdge(EdgeType_ et, GraphNode_ *t, GraphNode_ *s)
{
	GraphEdge_ * edge = (GraphEdge_ *) xmalloc (sizeof (*edge));
	edge->type = et;
	edge->target = t;
	edge->source = s;
	edge->next = 0;
	return edge;
}

static void free_edges()
{
	GraphEdge_ *c = edge_head;

	while (c)
	{
		edge_head = edge_head->next;
		free (c);
		c = edge_head;
	}
}

static GraphNode_ * find_graphNode(char * name)
{
	return hash_find_item (&graphNodes, GraphNode(NT_DIR, name)); 
}

static void insertGraphNode(GraphNode_ *node)
{
	GraphNode_ **node_slot;

	node_slot = (GraphNode_ **)hash_find_slot(&graphNodes, node);

	if (!HASH_VACANT(*node_slot))
	{
		assert(0);
	}
	
	hash_insert_at(&graphNodes, node, node_slot);
}

static GraphNode_ *get_graphNode(NodeType_ nt, const char *name)
{
	GraphNode_ *key = GraphNode(nt, name);
	GraphNode_ *res;
	res = hash_find_item (&graphNodes, key);
	if (res)
	{
		free_graphNode(key);
		return res;
	}

	insertGraphNode(key);
	return key;
}

/*
 * parse the information from BUF into LINE CONTENT,
 * LINE_CONTENT is created in this function
 * return 1 if succeed, otherwise return 0
 * */
static void analysis(const char *op, const char *opa1, const char *opa2)
{
	GraphNode_ *node, *node2;
	GraphEdge_ *edge;
	if (strcmp(op, "FILE") == 0)
       	{
		node = get_graphNode(NT_FILE, opa1);
		assert (!node->filename);
		assert (strlen(opa2) > 0);
		node->type = NT_FILE;
		node->filename = xstrdup(opa2);
	}
	
	else if (strcmp(op, "PHONY") == 0)
	{
		node = get_graphNode(NT_FILE, opa1);
		node->type = NT_FILE;
		node->isphony = 1;
	}

	else if (strcmp(op, "TARGET") == 0)
	{
		node = get_graphNode(NT_FILE, opa1);
		node->type = NT_FILE;
		node->istarget = 1;
	}

	else if (strcmp(op, "GOAL") == 0)
	{
		node = get_graphNode(NT_FILE, opa1);
		node->type = NT_FILE;
		node->isgoal = 1;
	}
	
	else if (strcmp(op, "EXECUTED") == 0)
	{
		node = get_graphNode(NT_FILE, opa1);
		node->type = NT_FILE;
		node->isexecuted = 1;
	}
	
	else if (strcmp(op, "DBLCLN") == 0)
	{
		node = get_graphNode(NT_FILE, opa1);
		node->type = NT_FILE;
		node->dblcln = atoi(opa2);
	}
	
	else if (strcmp(op, "CONTAIN") == 0)
	{
		node = get_graphNode(NT_DIR, opa1);
		node2= get_graphNode(NT_DIR, opa2);
		node2->next = node->children;
		node->children = node2;
	}
	
	else if (strcmp(op, "INVOKE") == 0)
	{
		node  = get_graphNode(NT_DIR, opa1);
		node2 = get_graphNode(NT_DIR, opa2);
		edge = GraphEdge(ET_INVOKE, node , node2);
		edge->next = edge_head;
		edge_head = edge;
	}
	
	else if (strcmp(op, "DEPENDON") == 0)
	{
		node  = get_graphNode(NT_DIR, opa1);
		node2 = get_graphNode(NT_DIR, opa2);
		edge = GraphEdge(ET_DEPENDON, node , node2);
		edge->next = edge_head;
		edge_head = edge;
	}
	
	else if (strcmp(op, "ROOT") == 0)
	{
		if (root_directory)  free (root_directory);
		root_directory = xstrdup(opa1);
	}
	
	else if (strcmp(op, "LAYERNAME") == 0)
	{
		layer_name[atoi(opa2)] = xstrdup(opa1);
	}
	
	else if (strcmp(op, "LAYER") == 0)
	{
		node = get_graphNode(NT_DIR, opa1);
		if (!node->layer)
			node->layer = xstrdup(opa2);
		else
		{
			free (node->layer);
			node->layer = xstrdup("all");
		}
	}
	
	else if (strcmp(op, "SAME") == 0)
	{
		node = get_graphNode(NT_DIR, opa1);
		node2 = get_graphNode(NT_DIR, opa2);
		if (node != node2)
		{
			node2->isdup = 1;
			node->same = node2;
			node2->same = node;
		}
	}
	
	else if (strcmp(op, "DIFF") == 0)
	{
		node = get_graphNode(NT_DIR, opa1);
		node->isdiff = 1;
	}
	
	else
	{
		printf("ERROR: unrecognized relation: %s\n", op);
		assert (0);
	}
}

static void print_file_header(FILE *fp)
{
	/* output the head info of digraph */
        fprintf(fp, "digraph G {\n");
        fprintf(fp, "\tsize = \"10,7.5\"; ratio = fill; rotate = 90\n");
	if (layer_name[0] && layer_name[1])
	{
		/*
		fprintf(fp, "\tlayers = \"%s:%s\"\n", 
				layer_name[0], layer_name[1]);
				*/
	}
}

static void print_file_footer(FILE *fp)
{
	fprintf(fp, "}\n");
}

static void print_SubGraphHeader(int depth, char *name, FILE *fp)
{
	char * tmp_front, * tmp, * tmp_label;
	int size, i;

	tmp_front = (char *)xmalloc(sizeof(char) * (depth +1));
	strcpy(tmp_front, "\t");
	for (i = 1; i < depth; i++)
		strcat(tmp_front, "\t");
	tmp_front[depth] = 0;
	
	size = strlen(tmp_front) + strlen(name) + 22;
	tmp = (char *)xmalloc(sizeof(char) * (size + 1));
	strcpy(tmp, tmp_front);
	strcat(tmp, "subgraph \"cluster_" );
	strcat(tmp, name);
	strcat(tmp, "\" {\n");
	tmp[size] = 0;
        fwrite(tmp, strlen(tmp), 1, fp);
	free(tmp);

	size = strlen(tmp_front) + strlen(name) + 11;
	tmp = (char *)xmalloc(sizeof(char) * (size + 1));
	strcpy(tmp, tmp_front);
	strcat(tmp, "\tlabel=\"" );
	tmp_label = (char *)strrchr(name, '/');
	if (tmp_label && (strlen(tmp_label) > 1))
		strcat(tmp, (tmp_label + 1));
	else
		strcat(tmp, name);
	strcat(tmp, "\";\n");
	tmp[size] = 0;
        fwrite(tmp, strlen(tmp), 1, fp);
	free(tmp);
	free(tmp_front);
}

static void print_SubGraphFooter(int depth, FILE * fp)
{
	char *tmp_front;
	int i;
	
	tmp_front = (char *)xmalloc(sizeof(char) * (depth + 3));
	strcpy(tmp_front, "\t");
	for (i = 1; i < depth; i++)
		strcat(tmp_front, "\t");
	strcat(tmp_front, "}\n");
	tmp_front[depth+2] = 0;
        fwrite(tmp_front, strlen(tmp_front), 1, fp);
	free(tmp_front);
}

static void print_GraphNode(GraphNode_ *node, int depth, FILE *fp)
{
	char *prefix;
	int i;
	
	if (node == 0) return;

	if (node->isdup) 
	{
		print_GraphNode(node->next, depth, fp);
		return;
	}
	else if (node->children != 0)
	{
		print_SubGraphHeader(depth, node->name, fp);
		print_GraphNode(node->children, depth + 1, fp);
		if (node->same)
			print_GraphNode(node->same->children, depth + 1, fp);
		print_SubGraphFooter(depth, fp);
		print_GraphNode(node->next, depth, fp);
		return;
	}

	prefix = (char *)xmalloc(sizeof(char) * (depth + 1));
	strcpy(prefix, "\t");
	for (i = 1; i < depth; i++) strcat(prefix, "\t");
	prefix[depth] = 0;

	/* same */
	if (node->same)
		fprintf(fp, "%s\"%s %s\"", 
				prefix, node->name, node->same->name);
	else
		fprintf(fp, "%s\"%s\"", prefix, node->name);

	switch (node->type) 
	{
	case NT_DIR:
		fprintf(fp, "[shape=box,label=\"%s\"", node->filename);
		break;
			
	case NT_FILE:
		/* double colon */
		if (node->dblcln < 0)
			fprintf(fp, "[label=\"%s\"", node->filename);
		else
			fprintf(fp, "[label=\"%s ((%d))\"",
					node->filename, node->dblcln);
		/* goal */
		if (node->isgoal)
		{
			fprintf(fp, ",shape=doublecircle");
			/*fprintf(fp, ",shape=ellipse,peripheries=2");*/
			/*fprintf(fp, ",peripheries=2");*/
			/*fprintf(fp, ",shape=ellipse");*/
		}
		else
			fprintf(fp, ",shape=ellipse");
		/* phony */
		if (node->isphony)
			fprintf(fp, ",fontname=\"Courier\"");
		else
			fprintf(fp, ",fontname=\"Times-Roman\"");
		/* target */
		if (node->istarget)
			fprintf(fp, ",fontcolor=red");
			/*fprintf(fp, ",style=bold");*/
		else
			fprintf(fp, ",fontcolor=blue");
			/*fprintf(fp, ",style=dotted");*/
		/* executed */
		if (node->isexecuted)
			fprintf(fp, ",color=lightyellow,style=filled");
		break;

	default:
		assert(0);
	}

	/* layer */
	if (node->layer)
	{
		if (node->same)
			fprintf(fp, ",layer=\"all\"");
		else
			fprintf(fp, ",layer=\"%s\"", node->layer);
	}
	/* diff */
	if (node->isdiff)
		fprintf(fp, ",color=orange");
		/*fprintf(fp, ",shape=diamond");*/
	else if (node->same)
		fprintf(fp, ",color=black");
		/*fprintf(fp, ",shape=ellipse");*/
	else
	{
		if (node->layer && layer_name[0]
				&& strcmp(node->layer,layer_name[0])==0)
			fprintf(fp, ",color=red");
			/*fprintf(fp, ",shape=trapezium");*/
		else if (node->layer && layer_name[1]
				&& strcmp(node->layer,layer_name[1])==0)
			fprintf(fp, ",color=green");
			/*fprintf(fp, ",shape=invtrapezium");*/
	}

	fprintf(fp, "];\n");

	free (prefix);

	print_GraphNode(node->next, depth, fp);
}

static void print_components(FILE *fp)
{
	GraphNode_ *rootNode, *node;
	
	if (root_directory)
		rootNode = find_graphNode(root_directory);
	else
		rootNode = find_graphNode("/");

	print_GraphNode(rootNode, 1, fp);
}

static int isduplicated(GraphEdge_ *e)
{
	if (e->target->isdup && e->source->isdup)
		return 1;
	if (e->source->layer && strcmp(e->source->layer, "all") == 0
			&& e->target->isdup)
		return 1;
	if (e->target->layer && strcmp(e->target->layer, "all") == 0
			&& e->source->isdup)
		return 1;
	return 0;
}

static void print_dependency(GraphEdge_ *e, FILE *fp)
{
	char *layer_t = 0, *layer_s = 0;
	
	if (isduplicated(e)) return;
	
	fprintf(fp, "\t");
	
	if (!e->target->same)
	{
		layer_t = e->target->layer;
		fprintf(fp, "\"%s\"", e->target->name);
	}
	else if (!e->target->isdup)
		fprintf(fp,"\"%s %s\"", e->target->name, e->target->same->name);
	else
		fprintf(fp,"\"%s %s\"", e->target->same->name, e->target->name);

	fprintf(fp, "->");
	
	if (!e->source->same)
	{
		layer_s = e->source->layer;
		fprintf(fp, "\"%s\"", e->source->name);
	}
	else if (!e->source->isdup)
		fprintf(fp,"\"%s %s\"", e->source->name, e->source->same->name);
	else
		fprintf(fp,"\"%s %s\"", e->source->same->name, e->source->name);

	if (e->target->same && e->source->same)
		fprintf(fp, "[layer=\"all\",");
	
	else if (!layer_t && !layer_s)
		fprintf(fp, "[");
	
	else if (layer_t && !layer_s)
		fprintf(fp, "[layer=\"%s\",", layer_t);
	
	else if (!layer_t && layer_s)
		fprintf(fp, "[layer=\"%s\",", layer_s);
	
	else if (layer_t && layer_s && (strcmp(layer_t, layer_s) == 0))
		fprintf(fp, "[layer=\"%s\",", layer_t);
	else
	{
		printf("Error: conflict layer information\n");
		abort();
	}

	switch (e->type)
	{
		case ET_INVOKE:
			/*fprintf(fp, "color=blue];\n");*/
			fprintf(fp, "style=dotted];\n");
			break;
			
		case ET_DEPENDON:
			/*fprintf(fp, "color=black];\n");*/
			fprintf(fp, "style=filled];\n");
			break;
			
		default:
			assert(0);
	}
}

static void print_all_edges(FILE *fp)
{
	GraphEdge_ *e;
	for (e = edge_head; e; e = e->next) 
		print_dependency(e, fp);
}

/*
 * convert input file into dot file
 * return 0 if succeed, otherwise return 1
 * */
int convert_to_format(char *infilenm, char *outfilenm) {
	
	FILE * infp, * outfp;
	char op[32], opa1[1024], opa2[1024];
	int i;
	
	/* open input file */
	infp = fopen(infilenm, "r");
	if (infp == 0) 
	{
		printf("Error: can not open %s !!!\n", infilenm);
		abort();
	}

	/* open output file */
	outfp = fopen(outfilenm, "w");
	if (outfp == 0) 
	{
		printf("Error: can not open %s !!!\n", outfilenm);
		fclose (infp);
		abort();
	}

	hash_init (&graphNodes, 500, graph_hash_1, graph_hash_2, graph_hash_cmp);
	while ( fscanf(infp, "%s\t%s\t%s", op, opa1, opa2) != EOF)
		analysis(op, opa1, opa2);

	/* output the head info of digraph */
	print_file_header(outfp);

	print_components(outfp);
	print_all_edges(outfp);
	/* output the final part of the dot file */
	print_file_footer(outfp);
	
	free_hashtable(graphNodes);
	free_edges();
	if (root_directory)  free (root_directory);
	if (layer_name[0])   free (layer_name[0]);
	if (layer_name[1])   free (layer_name[1]);
	fclose (infp);
	fclose (outfp);
	return 0;
}

int
main(int argc, char ** argv, char **envp)
{
	int result;

	if (argc <= 1) 
	{
	       print_usage();	
	       return 1;
	}
	
	if (argc == 2) 
	{
		result = convert_to_format(argv[1], "output.dot"); 
	}
	else 
	{
		result = convert_to_format(argv[1], argv[2]);
	}

	if (result)
	{
		printf("\n Convert to DOT format failed. \n");
	}
	else
	{
		printf("\n Convert to DOT format succeeded. \n");
	}
	return result;
}
